"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Parser = void 0;
require("source-map-support/register");
const fetch = require("isomorphic-fetch");
const lodash_1 = require("lodash");
const magic_string_1 = require("magic-string");
const memoize_lit_1 = require("memoize-lit");
const parser_1 = require("./compiler/parser");
const tokenizer_1 = require("./compiler/tokenizer");
const transformer_1 = require("./compiler/transformer");
const methods_1 = require("./methods");
const REPEAT_WHITELIST = [
    'query',
    'randomnum',
    'add',
    'incr',
    'arg',
    'randlist',
    'listpick',
    'time',
    'user',
    'touser',
    'userid',
];
class Parser {
    constructor(middleware = {}) {
        this.middleware = middleware;
        this._methods = methods_1.methods;
        if (middleware === null || middleware === void 0 ? void 0 : middleware.onServiceAPI) {
            this.middleware.onServiceAPI = (0, memoize_lit_1.memoize)(middleware.onServiceAPI, { maxAge: 60000 });
        }
        if (middleware === null || middleware === void 0 ? void 0 : middleware.onHandleMethod) {
            this.middleware.onHandleMethod = (0, memoize_lit_1.memoize)(middleware.onHandleMethod, { maxAge: 60000 });
        }
    }
    parse(message_1, settings_1, text_1) {
        return __awaiter(this, arguments, void 0, function* (message, settings, text, cache = []) {
            const token = (0, tokenizer_1.tokenizer)(text);
            const parsed = (0, parser_1.parser)(token);
            const transform = (0, transformer_1.transformer)(parsed);
            const original = new magic_string_1.default(text);
            let context = {
                cache: [],
                request: (0, memoize_lit_1.memoize)(fetch),
            };
            if (cache != null && cache.length > 0) {
                context.cache = cache;
            }
            for (let i = 0, length = transform.body.length; i < length; i++) {
                const part = transform.body[i];
                if (part.type === 'ExpressionStatement' && String(part.expression.callee.name).toLowerCase() === 'repeat' && part.expression.callee.type === 'Identifier') {
                    const res = yield this.handleRepeat(context, message, settings, part.expression);
                    if (res == null) {
                        continue;
                    }
                    original.overwrite(part.start, part.end, (0, lodash_1.trim)(res.toString()));
                    continue;
                }
                if (part.type !== 'ExpressionStatement' || methods_1.methods[part.expression.callee.name.toLowerCase()] === undefined) {
                    continue;
                }
                const res = yield this.run(context, message, settings, part.expression, false);
                if (res == null) {
                    continue;
                }
                original.overwrite(part.start, part.end, (0, lodash_1.trim)(res.toString()));
            }
            context = undefined;
            return (0, lodash_1.trim)(original.toString());
        });
    }
    parseWithOutContext(text_1) {
        return __awaiter(this, arguments, void 0, function* (text, messageContext = {}, settingsContext = {}, cache = []) {
            let message = Object.assign({ channel: {
                    id: '1',
                    name: 'SimpleParser',
                }, provider: 'simple-parser', message: {
                    args: text.split(/\s+/),
                    raw: text,
                }, user: {
                    id: '1',
                    name: 'SimpleParser',
                } }, messageContext);
            const settings = Object.assign({ timezone: 'America/New_York' }, settingsContext);
            return this.parse(message, settings, text, cache);
        });
    }
    addMethods(methods) {
        if (Array.isArray(methods)) {
            for (const m of methods) {
                if (this._methods[m.name]) {
                    throw new Error(`Method ${m.name} already exists`);
                }
                this._methods[m.name] = m.method;
            }
            return;
        }
        if (this._methods[methods.name]) {
            throw new Error(`Method ${methods.name} already exists`);
        }
        this._methods[methods.name] = methods.method;
    }
    run(context_1, message_1, settings_1, expr_1) {
        return __awaiter(this, arguments, void 0, function* (context, message, settings, expr, isRepeat = false) {
            if (expr.type === 'String') {
                return expr.value;
            }
            if (expr.arguments.length === 0) {
                return this.handle(context, message, settings, expr.callee.name, [], isRepeat);
            }
            const args = yield Promise.all(expr.arguments.map(arg => {
                var _a, _b;
                if (((_b = (_a = arg === null || arg === void 0 ? void 0 : arg.callee) === null || _a === void 0 ? void 0 : _a.name) === null || _b === void 0 ? void 0 : _b.toLowerCase()) === 'repeat') {
                    return this.handleRepeat(context, message, settings, arg);
                }
                return this.run(context, message, settings, arg, isRepeat);
            }));
            return this.handle(context, message, settings, expr.callee.name, args, isRepeat);
        });
    }
    handle(context_1, message_1, settings_1, text_1) {
        return __awaiter(this, arguments, void 0, function* (context, message, settings, text, args = [], isRepeat) {
            if (methods_1.methods[text.toLowerCase()] == null) {
                return;
            }
            if (isRepeat && !REPEAT_WHITELIST.includes(text.toLowerCase())) {
                return;
            }
            return methods_1.methods[text.toLowerCase()].call(this, message, settings, context, ...args);
        });
    }
    handleRepeat(context, message, settings, expr) {
        return __awaiter(this, void 0, void 0, function* () {
            const MAX_REPEATS = 5;
            let TO_REPEAT = MAX_REPEATS;
            const repeatArgs = [...expr.arguments].pop();
            if (repeatArgs.type === 'String') {
                const userValue = Number(repeatArgs.value);
                if (!isNaN(userValue) && userValue <= MAX_REPEATS) {
                    TO_REPEAT = userValue;
                }
            }
            else if (repeatArgs.type === 'CallExpression') {
                const res = yield this.run(context, message, settings, repeatArgs, true);
                const userValue = Number(res);
                if (!isNaN(userValue) && userValue <= MAX_REPEATS) {
                    TO_REPEAT = userValue;
                }
            }
            const results = [];
            for (let i = 0; i < TO_REPEAT; i++) {
                const res = yield this.run(context, message, settings, expr.arguments[0], true);
                if (res != null) {
                    results.push(res);
                }
            }
            return results.toString();
        });
    }
}
exports.Parser = Parser;
//# sourceMappingURL=index.js.map